@@ -81,8 +81,8 @@ def findClass(className):
     return cls
 
 
-INDENT = '    '
-HALF_INDENT = '  '
+INDENT = '  '
+HALF_INDENT = ' '
 
 PRIMITIVES = { 'boolean': 'Z',
                'byte': 'B',
@@ -95,8 +95,12 @@ PRIMITIVES = { 'boolean': 'Z',
                'void': 'V' }
 
 RESERVED = set(['delete', 'and', 'or', 'not', 'xor', 'union', 'register',
-                'const', 'bool', 'operator', 'typeof', 'asm',
-                'NULL', 'DOMAIN', 'IGNORE'])
+                'const', 'bool', 'operator', 'typeof', 'asm', 'mutable',
+                'inline', 'typedef', 'struct',
+                'NULL', 'DOMAIN', 'IGNORE', 'min', 'max', 'PREFIX'])
+
+RENAME_METHOD_SUFFIX = '_'
+RENAME_FIELD_SUFFIX = '__'
 
 def cppname(name):
 
@@ -113,7 +117,10 @@ def cppnames(names):
 
 def absname(names):
 
-    return "::%s" %('::'.join(names))
+    if names:
+        return "::%s" %('::'.join(names))
+
+    return ''
 
 
 def typename(cls, current, const):
@@ -227,10 +234,13 @@ def addRequiredTypes(cls, typeset, generics):
         if Class.instance_(cls):
             cls = Class.cast_(cls)
             if not (cls.isPrimitive() or cls in typeset):
-                typeset.add(cls)
-                cls = cls.getGenericSuperclass()
-                if cls is not None:
-                    addRequiredTypes(cls, typeset, True)
+                if cls.isArray():
+                    addRequiredTypes(cls.getComponentType(), typeset, True)
+                else:
+                    typeset.add(cls)
+                    cls = cls.getGenericSuperclass()
+                    if cls is not None:
+                        addRequiredTypes(cls, typeset, True)
         elif ParameterizedType.instance_(cls):
             pt = ParameterizedType.cast_(cls)
             addRequiredTypes(pt.getRawType(), typeset, True)
@@ -332,9 +342,11 @@ def expandjar(path):
 def jcc(args):
 
     classNames = set()
+    listedClassNames = set()
+    listedMethodNames = {}
     packages = set()
     jars = []
-    classpath = []
+    classpath = [_jcc.CLASSPATH]
     libpath = []
     vmargs = ['-Djava.awt.headless=true']
     moduleName = None
@@ -342,6 +354,7 @@ def jcc(args):
     build = False
     install = False
     recompile = False
+    egg_info = False
     output = 'build'
     debug = False
     excludes = []
@@ -349,6 +362,7 @@ def jcc(args):
     mappings = {}
     sequences = {}
     renames = {}
+    use_full_names = False
     env = None
     wrapperFiles = 1
     prefix = None
@@ -365,7 +379,9 @@ def jcc(args):
     arch = []
     resources = []
     imports = {}
-
+    extra_setup_args = []
+    initvm_args = {}
+    
     i = 1
     while i < len(args):
         arg = args[i]
@@ -391,6 +407,9 @@ def jcc(args):
             elif arg == '--vmarg':
                 i += 1
                 vmargs.append(args[i])
+            elif arg == '--maxheap':
+                i += 1
+                initvm_args['maxheap'] = args[i]
             elif arg == '--python':
                 from python import python, module
                 i += 1
@@ -407,6 +426,12 @@ def jcc(args):
             elif arg == '--compile':
                 from python import compile
                 recompile = True
+            elif arg == '--egg-info':
+                from python import compile
+                egg_info = True
+            elif arg == '--extra-setup-arg':
+                i += 1
+                extra_setup_args.append(args[i])
             elif arg == '--output':
                 i += 1
                 output = args[i]
@@ -428,6 +453,8 @@ def jcc(args):
                 i += 1
                 renames.update(dict([arg.split('=')
                                      for arg in args[i].split(',')]))
+            elif arg == '--use_full_names':
+                use_full_names = True
             elif arg == '--files':
                 i += 1
                 wrapperFiles = args[i]
@@ -478,14 +505,20 @@ def jcc(args):
             else:
                 raise ValueError, "Invalid argument: %s" %(arg)
         else:
+            if ':' in arg:
+                arg, method = arg.split(':', 1)
+                listedMethodNames.setdefault(arg, set()).add(method)
             classNames.add(arg)
+            listedClassNames.add(arg)
         i += 1
 
     if libpath:
         vmargs.append('-Djava.library.path=' + os.pathsep.join(libpath))
 
-    env = initVM(os.pathsep.join(classpath) or None,
-                 maxstack='512k', vmargs=' '.join(vmargs))
+    initvm_args['maxstack'] = '512k'
+    initvm_args['vmargs'] = vmargs
+
+    env = initVM(os.pathsep.join(classpath) or None, **initvm_args)
 
     typeset = set()
     excludes = set(excludes)
@@ -496,7 +529,7 @@ def jcc(args):
         else:
             raise ValueError, "--shared must be used when using --import"
 
-    if recompile or not build and (install or dist):
+    if recompile or not build and (install or dist or egg_info):
         if moduleName is None:
             raise ValueError, 'module name not specified (use --python)'
         else:
@@ -504,7 +537,8 @@ def jcc(args):
                     install, dist, debug, jars, version,
                     prefix, root, install_dir, home_dir, use_distutils,
                     shared, compiler, modules, wininst, find_jvm_dll,
-                    arch, generics, resources, imports)
+                    arch, generics, resources, imports, use_full_names,
+                    egg_info, extra_setup_args)
     else:
         if imports:
             def walk((include, importset), dirname, names):
@@ -547,7 +581,8 @@ def jcc(args):
             if className.split('$', 1)[0] in excludes or className in excludes:
                 continue
             cls = findClass(className.replace('.', '/'))
-            if Modifier.isPublic(cls.getModifiers()):
+            if (Modifier.isPublic(cls.getModifiers()) or
+                className in listedClassNames):
                 addRequiredTypes(cls, typeset, generics)
 
         _dll_export = ''
@@ -570,11 +605,20 @@ def jcc(args):
                 out_cpp = file(os.path.join(cppdir, fileName), 'w')
 
         done = set()
+        pythonNames = {}
         for importset in imports.itervalues():
             done.update(importset)
+            if moduleName:
+                for cls in importset:
+                    name = split_pkg(cls.getName(), '.')[-1]
+                    if not use_full_names:
+                        if name in pythonNames:
+                            raise ValueError, (cls, 'python class name already in use, use --rename', name, pythonNames[name])
+                        else:
+                            pythonNames[name] = cls
 
         todo = typeset - done
-	if allInOne and wrapperFiles > 1:
+        if allInOne and wrapperFiles > 1:
             classesPerFile = max(1, len(todo) / wrapperFiles)
         classCount = 0
         while todo:
@@ -592,22 +636,29 @@ def jcc(args):
                 line(out_h, 0, '#define %s_H', '_'.join(names))
 
                 (superCls, constructors, methods, protectedMethods,
-                 fields, instanceFields, declares) = \
+                 methodNames, fields, instanceFields, declares) = \
                     header(env, out_h, cls, typeset, packages, excludes,
-                           generics, _dll_export)
+                           generics, listedMethodNames.get(cls.getName(), ()),
+                           _dll_export)
 
                 if not allInOne:
                     out_cpp = file(fileName + '.cpp', 'w')
                 names, superNames = code(env, out_cpp,
                                          cls, superCls, constructors,
                                          methods, protectedMethods,
-                                         fields, instanceFields, 
+                                         methodNames, fields, instanceFields, 
                                          declares, typeset)
                 if moduleName:
+                    name = renames.get(className) or names[-1]
+                    if not use_full_names:
+                        if name in pythonNames:
+                            raise ValueError, (cls, 'python class name already in use, use --rename', name, pythonNames[name])
+                        else:
+                            pythonNames[name] = cls
                     python(env, out_h, out_cpp,
                            cls, superCls, names, superNames,
                            constructors, methods, protectedMethods,
-                           fields, instanceFields,
+                           methodNames, fields, instanceFields,
                            mappings.get(className), sequences.get(className),
                            renames.get(className),
                            declares, typeset, moduleName, generics,
@@ -622,9 +673,9 @@ def jcc(args):
                 elif wrapperFiles > 1:
                     if classCount >= classesPerFile:
                         out_cpp.close()
-	                fileCount += 1
-	                fileName = '__wrap%02d__.cpp' %(fileCount)
-	                out_cpp = file(os.path.join(cppdir, fileName), 'w')
+                        fileCount += 1
+                        fileName = '__wrap%02d__.cpp' %(fileCount)
+                        out_cpp = file(os.path.join(cppdir, fileName), 'w')
                         classCount = 0
                         
             done.update(todo)
@@ -636,17 +687,19 @@ def jcc(args):
         if moduleName:
             out = file(os.path.join(cppdir, moduleName) + '.cpp', 'w')
             module(out, allInOne, done, imports, cppdir, moduleName,
-                   shared, generics)
+                   shared, generics, use_full_names)
             out.close()
-            if build or install or dist:
+            if build or install or dist or egg_info:
                 compile(env, os.path.dirname(args[0]), output, moduleName,
                         install, dist, debug, jars, version,
                         prefix, root, install_dir, home_dir, use_distutils,
                         shared, compiler, modules, wininst, find_jvm_dll,
-                        arch, generics, resources, imports)
+                        arch, generics, resources, imports, use_full_names,
+                        egg_info, extra_setup_args)
 
 
-def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
+def header(env, out, cls, typeset, packages, excludes, generics,
+           listedMethodNames, _dll_export):
 
     names = cls.getName().split('.')
     superCls = cls.getSuperclass()
@@ -692,7 +745,15 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
     for constructor in cls.getDeclaredConstructors():
         if Modifier.isPublic(constructor.getModifiers()):
             if generics:
-                params = constructor.getGenericParameterTypes()
+                genericParams = constructor.getGenericParameterTypes()
+                params = constructor.getParameterTypes()
+                # It appears that the implicit instance-of-the-declaring-class
+                # parameter of a non-static inner class is missing from 
+                # getGenericParameterTypes()
+                if len(params) == len(genericParams) + 1:
+                    params[1:] = genericParams
+                else:
+                    params = genericParams
                 if len(params) == 1:
                     if params[0] == cls:
                         continue
@@ -715,8 +776,11 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
     methods = {}
     protectedMethods = []
     for method in cls.getDeclaredMethods():
+        if method.isSynthetic():
+            continue
         modifiers = method.getModifiers()
-        if Modifier.isPublic(modifiers):
+        if (Modifier.isPublic(modifiers) or
+            method.getName() in listedMethodNames):
             if generics:
                 returnType = method.getGenericReturnType()
             else:
@@ -725,6 +789,9 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
                          generics):
                 continue
             sig = "%s:%s" %(method.getName(), signature(method, True))
+            # Apparently, overridden clone() methods are still returned via
+            # getDeclaredMethods(), so keep the one with the more precise
+            # return type, equal to cls.
             if sig in methods and returnType != cls:
                 continue
             if generics:
@@ -739,27 +806,6 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
                 methods[sig] = method
         elif Modifier.isProtected(modifiers):
             protectedMethods.append(method)
-    for interface in interfaces:
-        for method in interface.getMethods():
-            sig = "%s:%s" %(method.getName(), signature(method, True))
-            if sig not in methods:
-                if generics:
-                    param = method.getGenericReturnType()
-                else:
-                    param = method.getReturnType()
-                if not known(param, typeset, declares, packages, excludes,
-                             generics):
-                    continue
-                if generics:
-                    params = method.getGenericParameterTypes()
-                else:
-                    params = method.getParameterTypes()
-                for param in params:
-                    if not known(param, typeset, declares, packages, excludes,
-                                 generics):
-                        break
-                else:
-                    methods[sig] = method
 
     def _compare(m0, m1):
         value = cmp(m0.getName(), m1.getName())
@@ -769,6 +815,7 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
 
     methods = methods.values()
     sort(methods, fn=_compare)
+    methodNames = set([cppname(method.getName()) for method in methods])
 
     for constructor in constructors:
         if generics:
@@ -864,13 +911,14 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
     line(out, indent, 'static jmethodID *mids$;');
     if instanceFields:
         line(out, indent, 'static jfieldID *fids$;');
-    line(out, indent, 'static jclass initializeClass();');
+    line(out, indent, 'static bool live$;');
+    line(out, indent, 'static jclass initializeClass(bool);');
     line(out)
 
     line(out, indent, 'explicit %s(jobject obj) : %s(obj) {',
          cppname(names[-1]), absname(cppnames(superNames)))
     line(out, indent + 1, 'if (obj != NULL)');
-    line(out, indent + 2, 'initializeClass();')
+    line(out, indent + 2, 'env->getClass(initializeClass);')
     line(out, indent, '}')
     line(out, indent, '%s(const %s& obj) : %s(obj) {}',
          cppname(names[-1]), cppname(names[-1]),
@@ -881,6 +929,9 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
         for field in fields:
             fieldType = field.getType()
             fieldName = cppname(field.getName())
+            if fieldName in methodNames:
+                print >>sys.stderr, "  Warning: renaming static variable '%s' on class %s to '%s%s' since it is shadowed by a method of same name." %(fieldName, '.'.join(names), fieldName, RENAME_FIELD_SUFFIX)
+                fieldName += RENAME_FIELD_SUFFIX
             if fieldType.isPrimitive():
                 line(out, indent, 'static %s %s;',
                      typename(fieldType, cls, False), fieldName)
@@ -932,11 +983,11 @@ def header(env, out, cls, typeset, packages, excludes, generics, _dll_export):
         line(out, indent, '}')
 
     return (superCls, constructors, methods, protectedMethods,
-            fields, instanceFields, declares)
+            methodNames, fields, instanceFields, declares)
 
 
 def code(env, out, cls, superCls, constructors, methods, protectedMethods,
-         fields, instanceFields, declares, typeset):
+         methodNames, fields, instanceFields, declares, typeset):
 
     className = cls.getName()
     names = className.split('.')
@@ -967,10 +1018,13 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
     line(out, indent, 'jmethodID *%s::mids$ = NULL;', cppname(names[-1]))
     if instanceFields:
         line(out, indent, 'jfieldID *%s::fids$ = NULL;', cppname(names[-1]))
+    line(out, indent, 'bool %s::live$ = false;', cppname(names[-1]))
 
     for field in fields:
         fieldType = field.getType()
         fieldName = cppname(field.getName())
+        if fieldName in methodNames:
+            fieldName += RENAME_FIELD_SUFFIX
         typeName = typename(fieldType, cls, False)
         if fieldType.isPrimitive():
             line(out, indent, '%s %s::%s = (%s) 0;',
@@ -980,11 +1034,13 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
                  typeName, cppname(names[-1]), fieldName)
 
     line(out)
-    line(out, indent, 'jclass %s::initializeClass()', cppname(names[-1]))
+    line(out, indent, 'jclass %s::initializeClass(bool getOnly)',
+         cppname(names[-1]))
     line(out, indent, '{')
-    line(out, indent + 1, 'if (!class$)')
+    line(out, indent + 1, 'if (getOnly)')
+    line(out, indent + 2, 'return (jclass) (live$ ? class$->this$ : NULL);')
+    line(out, indent + 1, 'if (class$ == NULL)')
     line(out, indent + 1, '{')
-    line(out)
     line(out, indent + 2, 'jclass cls = (jclass) env->findClass("%s");',
          className.replace('.', '/'))
 
@@ -1034,17 +1090,21 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
         for field in fields:
             fieldType = field.getType()
             fieldName = field.getName()
+            cppFieldName = cppname(fieldName)
+            if cppFieldName in methodNames:
+                cppFieldName += RENAME_FIELD_SUFFIX
             if fieldType.isPrimitive():
                 line(out, indent + 2,
                      '%s = env->getStatic%sField(cls, "%s");',
-                     cppname(fieldName), fieldType.getName().capitalize(),
+                     cppFieldName, fieldType.getName().capitalize(),
                      fieldName)
             else:
                 line(out, indent + 2,
                      '%s = new %s(env->getStaticObjectField(cls, "%s", "%s"));',
-                     cppname(fieldName), typename(fieldType, cls, False),
+                     cppFieldName, typename(fieldType, cls, False),
                      fieldName, signature(field))
 
+    line(out, indent + 2, "live$ = true;")
     line(out, indent + 1, '}')
     line(out, indent + 1, 'return (jclass) class$->this$;')
     line(out, indent, '}')
@@ -1078,6 +1138,7 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
             this = 'cls'
             midns = ''
             const = ''
+            sig = signature(method)
         else:
             isStatic = False
             if superMethod is not None:
@@ -1085,13 +1146,14 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
                 this = 'this$, (jclass) %s::class$->this$' %(absname(cppnames(superNames)))
                 declaringClass = superMethod.getDeclaringClass()
                 midns = '%s::' %(typename(declaringClass, cls, False))
+                sig = signature(superMethod)
             else:
                 qualifier = ''
                 this = 'this$'
                 midns = ''
+                sig = signature(method)
             const = ' const'
 
-        sig = signature(method)
         decls, args = argnames(params, cls)
 
         line(out)
@@ -1100,7 +1162,8 @@ def code(env, out, cls, superCls, constructors, methods, protectedMethods,
              cppname(methodName), decls, const)
         line(out, indent, '{')
         if isStatic:
-            line(out, indent + 1, 'jclass cls = initializeClass();');
+            line(out, indent + 1,
+                 'jclass cls = env->getClass(initializeClass);');
         if returnType.isPrimitive():
             line(out, indent + 1,
                  '%senv->call%s%sMethod(%s, %smids$[%smid_%s_%s]%s);',
