@@ -16,10 +16,10 @@ import os, sys, platform, shutil, _jcc
 from itertools import izip
 
 from cpp import PRIMITIVES, INDENT, HALF_INDENT
-from cpp import cppname, cppnames, typename
+from cpp import RENAME_METHOD_SUFFIX, RENAME_FIELD_SUFFIX
+from cpp import cppname, cppnames, absname, typename, findClass
 from cpp import line, signature, find_method, split_pkg, sort
 from cpp import Modifier, Class, Method
-from _jcc import findClass
 from config import INCLUDES, CFLAGS, DEBUG_CFLAGS, LFLAGS, IMPLIB_LFLAGS, \
     SHARED, VERSION as JCC_VER
 
@@ -142,7 +142,7 @@ def parseArgs(params, current, generics, genericParams=None):
             return ''
         if is_boxed(clsName):
             clsNames = clsName.split('.')
-            return ', &%s::PY_TYPE(%s)' %('::'.join(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
+            return ', &%s::PY_TYPE(%s)' %(absname(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
         return ', %s::initializeClass' %(typename(cls, current, False))
 
     def callarg(cls, i):
@@ -288,8 +288,11 @@ def returnValue(cls, returnType, value, genericRT=None, typeParams=None):
             clsArgs = []
             for clsArg in getActualTypeArguments(genericRT):
                 if Class.instance_(clsArg):
-                    clsNames = Class.cast_(clsArg).getName().split('.')
-                    clsArg = '&%s::PY_TYPE(%s)' %('::'.join(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
+                    clsArg = Class.cast_(clsArg)
+                    if clsArg.isArray():
+                        clsArg = Class.forName('java.lang.Object')
+                    clsNames = clsArg.getName().split('.')
+                    clsArg = '&%s::PY_TYPE(%s)' %(absname(cppnames(clsNames[:-1])), cppname(clsNames[-1]))
                     clsArgs.append(clsArg)
                 elif TypeVariable.instance_(clsArg):
                     gd = TypeVariable.cast_(clsArg).getGenericDeclaration()
@@ -372,7 +375,7 @@ def call(out, indent, cls, inCase, method, names, cardinality, isExtension,
     name = cppname(name)
     if Modifier.isStatic(modifiers):
         line(out, indent, 'OBJ_CALL(%s%s::%s(%s));',
-             result, '::'.join(cppnames(names)), name,
+             result, absname(cppnames(names)), name,
              ', '.join(['a%d' %(i) for i in xrange(count)]))
     else:
         line(out, indent, 'OBJ_CALL(%sself->object.%s(%s));',
@@ -441,6 +444,22 @@ def jniargs(params):
     return ''
 
 
+def wrapper_typename(returnType, cls):
+
+    if returnType.isArray():
+        componentType = returnType.getComponentType()
+        componentName = typename(componentType, cls, False)
+        if componentType.isPrimitive():
+            return "t_JArrayWrapper< %s >" %(componentName)
+
+        ns, sep, n = rpartition(componentName, '::')
+        return "t_JArrayWrapper< jobject,%s%st_%s >" %(ns, sep, n)
+
+    returnName = typename(returnType, cls, False)
+    ns, sep, n = rpartition(returnName, '::')
+    return "%s%st_%s" %(ns, sep, n)
+
+
 def extension(env, out, indent, cls, names, name, count, method, generics):
 
     line(out, indent, 'jlong ptr = jenv->CallLongMethod(jobj, %s::mids$[%s::mid_pythonExtension_%s]);',
@@ -541,7 +560,8 @@ def extension(env, out, indent, cls, names, name, count, method, generics):
 
 
 def python(env, out_h, out, cls, superCls, names, superNames,
-           constructors, methods, protectedMethods, fields, instanceFields,
+           constructors, methods, protectedMethods,
+           methodNames, fields, instanceFields,
            mapping, sequence, rename, declares, typeset, moduleName, generics,
            _dll_export):
 
@@ -665,18 +685,18 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                     if name in allMethods:
                         if Modifier.isStatic(allMethods[name][0].getModifiers()):
                             allMethods[name].append(method)
-                        elif name + '_' in allMethods:
-                            allMethods[name + '_'].append(method)
+                        elif name + RENAME_METHOD_SUFFIX in allMethods:
+                            allMethods[name + RENAME_METHOD_SUFFIX].append(method)
                         else:
-                            print >>sys.stderr, "  Warning: renaming static method '%s' on class %s to '%s_' since it is shadowed by non-static method of same name." %(name, '.'.join(names), name)
-                            allMethods[name + '_'] = [method]
+                            print >>sys.stderr, "  Warning: renaming static method '%s' on class %s to '%s%s' since it is shadowed by non-static method of same name." %(name, '.'.join(names), name, RENAME_METHOD_SUFFIX)
+                            allMethods[name + RENAME_METHOD_SUFFIX] = [method]
                     else:
                         allMethods[name] = [method]
                 else:
                     if name in allMethods:
                         if Modifier.isStatic(allMethods[name][0].getModifiers()):
-                            print >>sys.stderr, "  Warning: renaming static method '%s' on class %s to '%s_' since it is shadowed by non-static method of same name." %(name, '.'.join(names), name)
-                            allMethods[name + '_'] = allMethods[name]
+                            print >>sys.stderr, "  Warning: renaming static method '%s' on class %s to '%s%s' since it is shadowed by non-static method of same name." %(name, '.'.join(names), name, RENAME_METHOD_SUFFIX)
+                            allMethods[name + RENAME_METHOD_SUFFIX] = allMethods[name]
                             allMethods[name] = [method]
                         else:
                             allMethods[name].append(method)
@@ -687,14 +707,16 @@ def python(env, out_h, out, cls, superCls, names, superNames,
                 nameLen = len(name)
                 paramsLen = len(params)
                 if nameLen > 3 and paramsLen == 0 and name.startswith('get'):
-                    propMethods.setdefault(name[3].lower() + name[4:],
-                                           []).append(method)
+                    if method.getReturnType().getName() != 'void':
+                        propMethods.setdefault(name[3].lower() + name[4:],
+                                               []).append(method)
                 elif nameLen > 3 and paramsLen == 1 and name.startswith('set'):
                     propMethods.setdefault(name[3].lower() + name[4:],
                                            []).append(method)
                 elif nameLen > 2 and paramsLen == 0 and name.startswith('is'):
-                    propMethods.setdefault(name[2].lower() + name[3:],
-                                           []).append(method)
+                    if method.getReturnType().getName() != 'void':
+                        propMethods.setdefault(name[2].lower() + name[3:],
+                                               []).append(method)
 
     properties = set([name for name in propMethods.iterkeys()
                       if name not in allMethods])
@@ -903,16 +925,14 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         if iteratorExt:
             tp_iter = 'get_extension_iterator'
         else:
-            tp_iter = '((PyObject *(*)(t_%s *)) get_%siterator<t_%s>)' %(names[-1], clsParams and 'generic_' or '', names[-1])
+            tp_iter = '((PyObject *(*)(t_%s *)) get_%siterator< t_%s >)' %(names[-1], clsParams and 'generic_' or '', names[-1])
         tp_iternext = '0'
     elif nextMethod and iterable is not None and iterator.isAssignableFrom(cls):
         tp_iter = 'PyObject_SelfIter'
-        returnName = typename(nextMethod.getReturnType(), cls, False)
-        ns, sep, n = rpartition(returnName, '::')
         if nextExt:
             tp_iternext = 'get_extension_next'
         else:
-            tp_iternext = '((PyObject *(*)(java::util::t_Iterator *)) get_%siterator_next<java::util::t_Iterator,%s%st_%s>)' %(clsParams and 'generic_' or '', ns, sep, n)
+            tp_iternext = '((PyObject *(*)(::java::util::t_Iterator *)) get_%siterator_next< ::java::util::t_Iterator,%s >)' %(clsParams and 'generic_' or '', wrapper_typename(nextMethod.getReturnType(), cls))
     elif nextElementMethod and enumeration.isAssignableFrom(cls):
         tp_iter = 'PyObject_SelfIter'
         returnName = typename(nextElementMethod.getReturnType(), cls, False)
@@ -920,15 +940,14 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         if nextElementExt:
             tp_iternext = 'get_extension_nextElement'
         else:
-            tp_iternext = '((PyObject *(*)(java::util::t_Enumeration *)) get_%senumeration_next<java::util::t_Enumeration,%s%st_%s>)' %(clsParams and 'generic_' or '', ns, sep, n)
+            tp_iternext = '((PyObject *(*)(::java::util::t_Enumeration *)) get_%senumeration_next< ::java::util::t_Enumeration,%s >)' %(clsParams and 'generic_' or '', wrapper_typename(nextElementMethod.getReturnType(), cls))
     elif nextMethod:
         tp_iter = 'PyObject_SelfIter'
-        returnName = typename(nextMethod.getReturnType(), cls, False)
-        ns, sep, n = rpartition(returnName, '::')
         if nextExt:
             tp_iternext = 'get_extension_next'
         else:
-            tp_iternext = '((PyObject *(*)(t_%s *)) get_%snext<t_%s,%s%st_%s,%s>)' %(names[-1], clsParams and 'generic_' or '', names[-1], ns, sep, n, returnName)
+            returnType = nextMethod.getReturnType()
+            tp_iternext = '((PyObject *(*)(t_%s *)) get_%snext< t_%s,%s,%s >)' %(names[-1], clsParams and 'generic_' or '', names[-1], wrapper_typename(returnType, cls), typename(returnType, cls, False))
     else:
         tp_iter = '0'
         tp_iternext = '0'
@@ -988,7 +1007,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         tp_as_sequence = '0'
 
     if len(superNames) > 1:
-        base = '::'.join(('::'.join(cppnames(superNames[:-1])), superNames[-1]))
+        base = '::'.join((absname(cppnames(superNames[:-1])), superNames[-1]))
     else:
         base = superNames[-1]
     line(out)
@@ -1069,10 +1088,10 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "boxfn_", make_descriptor(%s));', names[-1], boxfn_)
 
     if isExtension:
-        line(out, indent + 1, 'jclass cls = %s::initializeClass();',
+        line(out, indent + 1, 'jclass cls = env->getClass(%s::initializeClass);',
              cppname(names[-1]))
     elif fields:
-        line(out, indent + 1, '%s::initializeClass();', cppname(names[-1]))
+        line(out, indent + 1, 'env->getClass(%s::initializeClass);', cppname(names[-1]))
 
     if isExtension:
         count = 0
@@ -1089,7 +1108,11 @@ def python(env, out_h, out, cls, superCls, names, superNames,
     for field in fields:
         fieldType = field.getType()
         fieldName = field.getName()
-        value = '%s::%s' %(cppname(names[-1]), cppname(fieldName))
+        cppFieldName = cppname(fieldName)
+        if cppFieldName in methodNames:
+            fieldName += RENAME_FIELD_SUFFIX
+            cppFieldName += RENAME_FIELD_SUFFIX
+        value = '%s::%s' %(cppname(names[-1]), cppFieldName)
         value = fieldValue(cls, value, fieldType)
         line(out, indent + 1, 'PyDict_SetItemString(PY_TYPE(%s).tp_dict, "%s", make_descriptor(%s));',
              names[-1], fieldName, value)
@@ -1401,7 +1424,7 @@ def python(env, out_h, out, cls, superCls, names, superNames,
         line(out, indent, '}')
 
 
-def package(out, allInOne, cppdir, namespace, names):
+def package(out, allInOne, cppdir, namespace, names, use_full_names):
 
     if not allInOne:
         out = file(os.path.join(os.path.join(cppdir, *names),
@@ -1416,6 +1439,8 @@ def package(out, allInOne, cppdir, namespace, names):
     if not names:
         line(out)
         line(out, 0, 'PyObject *initVM(PyObject *module, PyObject *args, PyObject *kwds);')
+        if use_full_names:
+            line(out, 0, 'PyObject *getJavaModule(PyObject *module, const char *parent, const char *name);')
 
     packages = []
     types = []
@@ -1449,6 +1474,13 @@ def package(out, allInOne, cppdir, namespace, names):
     line(out)
     line(out, indent, 'void __install__(PyObject *module)')
     line(out, indent, '{')
+
+    if use_full_names and names:
+        parent_name = '.'.join(names[:-1])
+        line(out, indent + 1, 'module = getJavaModule(module, "%s", "%s");',
+             parent_name, names[-1])
+        line(out)
+
     for name in types:
         line(out, indent + 1, 't_%s::install(module);', name)
     for name, entries in packages:
@@ -1469,6 +1501,12 @@ def package(out, allInOne, cppdir, namespace, names):
     else:
         line(out, indent, 'void __initialize__(PyObject *module)')
         line(out, indent, '{')
+
+    if use_full_names and names:
+        line(out, indent + 1, 'module = getJavaModule(module, "%s", "%s");',
+             parent_name, names[-1])
+        line(out)
+
     for name in types:
         line(out, indent + 1, 't_%s::initialize(module);', name)
     for name, entries in packages:
@@ -1497,11 +1535,11 @@ def package(out, allInOne, cppdir, namespace, names):
         line(out)
 
     for name, entries in packages:
-        package(out, allInOne, cppdir, entries, names + (name,))
+        package(out, allInOne, cppdir, entries, names + (name,), use_full_names)
 
 
 def module(out, allInOne, classes, imports, cppdir, moduleName,
-           shared, generics):
+           shared, generics, use_full_names):
 
     extname = '_%s' %(moduleName)
     line(out, 0, '#include <Python.h>')
@@ -1522,10 +1560,10 @@ def module(out, allInOne, classes, imports, cppdir, moduleName,
                 namespace = namespace.setdefault(className, {})
             namespace[classNames[-1]] = True
     if allInOne:
-        package(out_init, True, cppdir, namespaces, ())
+        package(out_init, True, cppdir, namespaces, (), use_full_names)
         out_init.close()
     else:
-        package(None, False, cppdir, namespaces, ())
+        package(None, False, cppdir, namespaces, (), use_full_names)
 
     line(out)
     line(out, 0, 'PyObject *initJCC(PyObject *module);')
@@ -1556,8 +1594,7 @@ def module(out, allInOne, classes, imports, cppdir, moduleName,
 def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
             version, prefix, root, install_dir, home_dir, use_distutils,
             shared, compiler, modules, wininst, find_jvm_dll, arch, generics,
-            resources, imports):
-
+            resources, imports, use_full_names, egg_info, extra_setup_args):
     try:
         if use_distutils:
             raise ImportError
@@ -1621,12 +1658,13 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
             copytree(resource, os.path.split(resource)[-1])
 
     packages = [moduleName]
+    package = [moduleName]
     if modules:
         for module in modules:
             if os.path.isdir(module):
                 def copytree(src, dst, is_package):
                     if is_package:
-                        packages.append('.'.join((moduleName, src.replace(os.path.sep, '.'))))
+                        packages.append('.'.join(package))
                     if not os.path.exists(dst):
                         os.mkdir(dst)
                     for name in os.listdir(src):
@@ -1637,13 +1675,17 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
                             continue
                         _dst = os.path.join(dst, name)
                         if os.path.isdir(_src):
+                            package.append(os.path.basename(_src))
                             copytree(_src, _dst, os.path.exists(os.path.join(_src, '__init__.py')))
+                            package.pop()
                         elif not is_package or name.endswith('.py'):
                             shutil.copy2(_src, _dst)
                 dst = modulePath
                 if os.path.exists(os.path.join(module, '__init__.py')):
                     dst = os.path.join(modulePath, os.path.basename(module))
+                    package.append(os.path.basename(module))
                     copytree(module, dst, True)
+                    package.pop()
                 else:
                     copytree(module, dst, False)
             else:
@@ -1678,6 +1720,8 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
     for import_ in imports:
         line(out, 0, 'from %s._%s import *', import_.__name__, import_.__name__)
     line(out, 0, 'from %s import *', extname)
+    if use_full_names:
+        line(out, 0, 'from java.io import PrintWriter, StringWriter')
     out.close()
 
     includes = [os.path.join(output, extname),
@@ -1690,7 +1734,7 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
         sources.append('jcc.cpp')
         sources.append('JCCEnv.cpp')
     for source in sources:
-	shutil.copy2(os.path.join(jccPath, 'sources', source),
+        shutil.copy2(os.path.join(jccPath, 'sources', source),
                      os.path.join(output, extname))
 
     if shared:
@@ -1718,7 +1762,10 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
             if name.endswith('.cpp'):
                 sources.append(os.path.join(path, name))
 
-    script_args = ['build_ext']
+    if egg_info:
+        script_args = ['egg_info']
+    else:    
+        script_args = ['build_ext']
 
     includes[0:0] = INCLUDES
     compile_args = CFLAGS
@@ -1738,7 +1785,7 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
         script_args.append('--debug')
         compile_args += DEBUG_CFLAGS
     elif sys.platform == 'win32':
-	pass
+        pass
     elif sys.platform == 'sunos5':
         link_args.append('-Wl,-s')
     else:
@@ -1828,6 +1875,7 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
         config_vars['CFLAGS'] = ' '.join(cflags)
 
     extensions = [Extension('.'.join([moduleName, extname]), **args)]
+    script_args.extend(extra_setup_args)
 
     args = {
         'name': moduleName,
@@ -1841,4 +1889,6 @@ def compile(env, jccPath, output, moduleName, install, dist, debug, jars,
     if with_setuptools:
         args['zip_safe'] = False
 
+    print "setup args = %s" % args
+
     setup(**args)
